
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_SERIALIZABLE_HH_
#define _MOOF_SERIALIZABLE_HH_

#include <iostream>
#include <map>
#include <string>
#include <vector>

#include <Moof/Deserializer.hh>
#include <Moof/Serializer.hh>
#include <Moof/StringTools.hh>


namespace Mf {


/**
 * Interface for a type which can be serialized and deserialized.
 */

class Serializable
{
public:
	virtual ~Serializable();

	virtual void serialize(Serializer& out) const = 0;
	virtual void deserialize(Deserializer& in) = 0;

	virtual void print() const = 0;

	virtual bool get(long& value);
	virtual bool get(double& value);
	virtual bool get(bool& value);
	virtual bool get(std::string& value);
	virtual bool get(std::wstring& value);
	virtual bool get(std::vector<SerializablePtr>& value);
	virtual bool get(std::map<std::string,SerializablePtr>& value);

	/*
	 * To get a number value which may have been parsed as either an integer or
	 * double, use these getters instead.
	 */

	bool getNumber(long&);
	bool getNumber(double&);

	virtual bool isNull();
};


template <class T>
class SerializableBase : public Serializable
{
public:
	SerializableBase() {}
	SerializableBase(const T& value) :
		value_(value) {}

	void serialize(Serializer& out) const;
	void deserialize(Deserializer& in);

	void print() const;
	bool get(T& value);
public:
	T value_;
};


class SerializableNull : public Serializable
{
public:
	SerializableNull() {}
	void serialize(Serializer& out) const;
	void deserialize(Deserializer& in);
	
	void print() const;
	bool isNull();
};


typedef SerializableBase<long>								SerializableInteger;
typedef SerializableBase<double>							SerializableReal;
typedef SerializableBase<bool>								SerializableBoolean;
typedef SerializableBase<std::string>						SerializableString;
typedef SerializableBase<std::wstring>						SerializableWideString;
typedef SerializableBase<std::vector<SerializablePtr> >	SerializableArray;
typedef SerializableBase<std::map<std::string,SerializablePtr> >
															SerializableMap;



template <class T>
inline void SerializableBase<T>::serialize(Serializer& out) const
{
	out.push(value_);
}

template <>
inline void
SerializableBase<std::vector<SerializablePtr> >::serialize(Serializer& out) const
{
	out.pushArrayHead();

	std::vector<SerializablePtr>::const_iterator it;
	for (it = value_.begin(); it < value_.end(); ++it)
	{
		(*it)->serialize(out);
	}

	out.pushArrayTail();
}

template <>
inline void
SerializableBase<std::map<std::string,SerializablePtr> >::serialize(Serializer& out) const
{
	out.pushMapHead();

	std::map<std::string,SerializablePtr>::const_iterator it;
	for (it = value_.begin(); it != value_.end(); ++it)
	{
		out.push((*it).first);
		(*it).second->serialize(out);
	}

	out.pushMapTail();
}

inline void SerializableNull::serialize(Serializer& out) const
{
	out.pushNull();
}


template <class T>
inline void SerializableBase<T>::deserialize(Deserializer& in)
{
	in.pop();
}

template <>
inline void SerializableBase<std::vector<SerializablePtr> >::deserialize(Deserializer& in)
{
	SerializablePtr obj;

	in.pop();

	while (obj = in.deserialize())
	{
		value_.push_back(SerializablePtr(obj));
	}

	in.pop();
}

template <>
inline void
SerializableBase<std::map<std::string,SerializablePtr> >::deserialize(Deserializer& in)
{
	SerializablePtr obj;

	in.pop();

	while (obj = in.deserialize())
	{
		std::string key;
		if (obj->get(key))
		{
			value_[key] = in.deserialize();
		}
	}

	in.pop();
}

inline void SerializableNull::deserialize(Deserializer& in)
{
	in.pop();
}


template <class T>
inline void SerializableBase<T>::print() const
{
	std::cout << std::boolalpha << typeid(T).name() << "(" << value_ << ")";
}

template <>
inline void SerializableBase<std::wstring>::print() const
{
	std::wcout << value_;
}

template <>
inline void SerializableBase<std::vector<SerializablePtr> >::print() const
{
	std::cout << "array";
}

template <>
inline void SerializableBase<std::map<std::string,SerializablePtr> >::print() const
{
	std::cout << "map";
}

inline void SerializableNull::print() const
{
	std::cout << "null";
}


template <class T>
inline bool SerializableBase<T>::get(T& value)
{
	value = value_;
	return true;
}

inline bool SerializableNull::isNull()
{
	return true;
}


} // namespace Mf

#endif // _MOOF_SERIALIZABLE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

